----------------------------------------------------------------------------------
-- Engineer:       Stanislav Maslan, s.maslan@seznam.cz
-- www:            http://www.elektronika.kvalitne.cz
--
-- Create Date:    15:08:21 05/01/2012 
-- Target Devices: CoolRunner2 XC2C256-7
-- Project Name:   640x480x16-bit TFT controller

-- Description:
-- ------------
-- Simple bitmap demo. CPLD is clocked at double rate than the TFT because I wasn't 
-- able to reach correct timing, so it requires 50MHz or so to reach recomended 
-- 25MHz pixel clock. 
----------------------------------------------------------------------------------
---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

-- other libs
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.numeric_std.all;


entity main is
	------ ports ------
	port(
		-- button input
		btn: in STD_LOGIC;
		-- source clock (50MHz)
		iclk: in STD_LOGIC;
		-- TFT pixel clock
		pclk: out STD_LOGIC;
		-- TFT synchronizations
		hsync: out STD_LOGIC;
		vsync: out STD_LOGIC;
		-- TFT color data
		red: out STD_LOGIC_VECTOR(4 downto 0);
		grn: out STD_LOGIC_VECTOR(5 downto 0);
		blu: out STD_LOGIC_VECTOR(4 downto 0));

	------ TFT definition constants ------
	-- TFT clocking parameters
	constant xw: integer := 790;
	constant yw: integer := 520;
	constant hsw: integer := 16;
	constant vsw: integer := 16;
	
	-- TFT visible area
	constant xofs: integer := 100;
	constant yofs: integer := 34;
	constant xsize: integer := 640;
	constant ysize: integer := 480;
	
	-- TFT color ranges
	constant rmax: integer := 31;
	constant gmax: integer := 63;
	constant bmax: integer := 31;
	
	--- sin/cos constants ---
	-- angle: integer bits
	constant sicos: integer := 5;
	-- angle: fraction bits
	constant sicof: integer := 3;
	-- angle: fraction modulo
	constant sicofn: integer := 8;
	-- sin/cos bits
	constant sicobit: integer := 10;
	-- sin/cos inititals
	constant sin0: integer := 0;
	constant cos0: integer := 808;
	
	-- chess-board bits
	constant chsbit: integer := 6;
end main;


architecture Behavioral of main is
	-- internal clock
	signal xclk: bit;
	-- mode switch
	signal pbtn: std_logic := '0';
	signal mode: bit := '0';	
	-- TFT position coutners
	signal pcnt: integer range 0 to xw;
	signal rcnt: integer range 0 to yw;
	-- color
	signal sred: integer range 0 to rmax := rmax;
	signal sgrn: integer range 0 to gmax := 0;
	signal sblu: integer range 0 to bmax := 0;
	-- chess-board generator
	signal xpos: integer range 0 to 2**(sicobit+chsbit)-1 := 0;
	signal ypos: integer range 0 to 2**(sicobit+chsbit)-1 := 0;
	signal xmem: integer range 0 to 2**(sicobit+chsbit)-1 := 0;
	signal ymem: integer range 0 to 2**(sicobit+chsbit)-1 := 0;
	-- sin/cos generator
	signal angf: integer range 0 to sicofn-1 := 0;
	signal ang: integer range 0 to 2**(1+sicos)-1 := 0;
	signal sin: integer range -2**sicobit to 2**sicobit-1 := sin0;
	signal cos: integer range -2**sicobit to 2**sicobit-1 := cos0;
begin
	
	process(iclk,xclk,btn,pbtn,pcnt,rcnt,sred,sgrn,sblu)
		variable pix: std_logic;
		variable fsync: boolean;
		variable angi: std_logic_vector(sicos downto 0);
		variable sdel: integer range -2**sicobit to 2**sicobit-1;
		variable cdel: integer range -2**sicobit to 2**sicobit-1;
	begin
		------ master clock divider ------
		if(iclk='0' and iclk'event) then
			-- internal clock signal
			xclk <= not xclk;
		end if;
			
		------ sync counters ------
		if(xclk='0' and xclk'event) then
			-- pixel counter
			if(pcnt>=xw) then
				pcnt <= 0;
				
				-- row counter
				if(rcnt>=yw) then
					rcnt <= 0;
				else
					rcnt <= rcnt + 1;
				end if;
			else
				pcnt <= pcnt + 1;
			end if;
		end if;
		
		------ internal frame sync signal
		fsync := pcnt=0 and rcnt=0;
				
		------ generate rainbow ------
		if(fsync and xclk='0' and xclk'event) then
			if(sred<rmax and sgrn=0 and sblu=bmax) then
				sred <= sred + 1;
			elsif(sred=rmax and sgrn<gmax and sblu=0) then
				sgrn <= sgrn + 1;
			elsif(sred=0 and sgrn=gmax and sblu<bmax) then
				sblu <= sblu + 1;
			elsif(sred>0 and sgrn=gmax and sblu=0) then
				sred <= sred - 1;
			elsif(sred=0 and sgrn>0 and sblu=bmax) then
				sgrn <= sgrn - 1;
			elsif(sred=rmax and sgrn=0 and sblu>0) then
				sblu <= sblu - 1;
			end if;
		end if;
		
		------ sine/cosine functions generator ------
		if(mode='0' and fsync and xclk='0' and xclk'event) then
			-- angle (integer part)
			angi := conv_std_logic_vector(ang,1+sicos);
			
			-- get sin/cos differences
			case conv_integer(angi(sicos-1 downto 0)) is
				when 0 => sdel := 11; cdel := 10;
				when 1 => sdel := 12; cdel := 7;
				when 2 => sdel := 14; cdel := 6;
				when 3 => sdel := 15; cdel := 2;
				when 4 => sdel := 15; cdel := 0;
				when 5 => sdel := 15; cdel := -3;
				when 6 => sdel := 13; cdel := -7;
				when 7 => sdel := 12; cdel := -9;
				when 8 => sdel := 9; cdel := -12;
				when 9 => sdel := 7; cdel := -13;
				when 10 => sdel := 3; cdel := -15;
				when 11 => sdel := 0; cdel := -15;
				when 12 => sdel := -2; cdel := -15;
				when 13 => sdel := -6; cdel := -14;
				when 14 => sdel := -7; cdel := -12;
				when 15 => sdel := -10; cdel := -11;
				when 16 => sdel := -10; cdel := -9;
				when 17 => sdel := -10; cdel := -7;
				when 18 => sdel := -11; cdel := -5;
				when 19 => sdel := -9; cdel := -4;
				when 20 => sdel := -8; cdel := -3;
				when 21 => sdel := -7; cdel := -3;
				when 22 => sdel := -6; cdel := -2;
				when 23 => sdel := -4; cdel := -3;
				when 24 => sdel := -3; cdel := -4;
				when 25 => sdel := -2; cdel := -6;
				when 26 => sdel := -3; cdel := -7;
				when 27 => sdel := -3; cdel := -8;
				when 28 => sdel := -4; cdel := -9;
				when 29 => sdel := -5; cdel := -11;
				when 30 => sdel := -7; cdel := -10;
				when others => sdel := -9; cdel := -10;
			end case;
			
			-- integrate sin/cos differences
			if(angi(sicos)='1') then
				sin <= sin - sdel;
				cos <= cos - cdel;
			else
				sin <= sin + sdel;
				cos <= cos + cdel;
			end if;
			
			-- lookup addressing coutners
			if(angf>=sicofn-1) then
				angf <= 0;
				ang <= ang + 1;
			else
				angf <= angf + 1;
			end if;
		end if;
		
		------ chess-board generator ------
		if(xclk='0' and xclk'event) then
			-- row coordiantes
			if(pcnt=0 and rcnt>=yofs) then
				xmem <= xmem + sin;
				ymem <= ymem - cos;
			elsif(pcnt=0) then
				xmem <= 0;
				ymem <= 0;
			end if;
			
			-- column coordinates
			if(pcnt>=xofs) then
				xpos <= xpos + cos;
				ypos <= ypos + sin;
			else
				xpos <= xmem;
				ypos <= ymem;
			end if;
			
			-- [col,row] to chess-board
			if(xpos>=2**(sicobit+chsbit-1) xor ypos>=2**(sicobit+chsbit-1)) then
				pix := '1';
			else
				pix := '0';
			end if;
		end if;
				
		------ put it out using rainbow color ------
		if(pix='1') then
			-- active pixel
			red <= conv_std_logic_vector(sred,red'length);
			grn <= conv_std_logic_vector(sgrn,grn'length);
			blu <= conv_std_logic_vector(sblu,blu'length);
		else
			-- inactive black
			red <= conv_std_logic_vector(0,red'length);
			grn <= conv_std_logic_vector(0,grn'length);
			blu <= conv_std_logic_vector(0,blu'length);		
		end if;
		
		------ background mode switch ------
		-- latch button state
		if(fsync and xclk='0' and xclk'event) then
			pbtn <= btn;
		end if;
		
		-- change mode
		if(pbtn='0' and pbtn'event) then
			mode <= not mode;
		end if;
		
	end process;
	
	------ TFT clocks generator ------
	-- pixel clock
	pclk <= '1' when iclk='0' and xclk='1' else '0';
	
	-- hsync
	hsync <= '1' when pcnt<(xw-hsw) else '0';
			
	-- vsync
	vsync <= '1' when rcnt>vsw else '0';
		
end Behavioral;

